/********************************************************************************
 * Copyright (C) 2018 TypeFox and others.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied: GNU General Public License, version 2
 * with the GNU Classpath Exception which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 ********************************************************************************/

import { injectable, inject, postConstruct } from 'inversify';
import { DisposableCollection, Emitter, Event, MessageService } from '@theia/core';
import { FileSystem, FileStat } from '@theia/filesystem/lib/common';
import { LocalStorageService } from '@theia/core/lib/browser';
import URI from '@theia/core/lib/common/uri';
import { WorkspaceService } from '@theia/workspace/lib/browser/workspace-service';
import { GitRepositoryTracker } from './git-repository-tracker';
import { ProblemManager } from '@theia/markers/lib/browser/problem/problem-manager';
import { Feature, GAD, FeatureVersion } from '../common';

import { AuthService } from '@quarks/auth-extension/lib/browser/auth-service';

export interface BuildEvent {
    type: 'start' | 'end',
    result?: 'ok' | 'nok',
    details?: any
}

export interface DeployEvent {
    type: 'start' | 'end',
    result?: 'ok' | 'nok',
    details?: any
}

@injectable()
export class GADService {

    protected _selectedFeature?: Feature;

    protected _workspaceFeatures?: Array<Feature>;

    protected toDispose = new DisposableCollection();

    protected readonly selectedFeatureStorageKey = 'quarks-selected-feature';

    protected readonly workspaceFeaturesStorageKey = 'quarks-workpace-features';

    protected readonly onFeatureChangeEmitter = new Emitter<Feature>();

    readonly onFeatureChange: Event<Feature> = this.onFeatureChangeEmitter.event;

    protected readonly onBuildEmitter = new Emitter<BuildEvent>();

    readonly onBuild: Event<BuildEvent> = this.onBuildEmitter.event;

    protected readonly onDeployEmitter = new Emitter<DeployEvent>();

    readonly onDeploy: Event<DeployEvent> = this.onDeployEmitter.event;

    protected deploying: boolean;

    protected building: boolean;

    isDeploying(): boolean {
        return this.deploying;
    }

    isBuilding(): boolean {
        return this.building;
    }

    constructor(
        @inject(GitRepositoryTracker) protected readonly gitRepositoryTracker: GitRepositoryTracker,
        @inject(LocalStorageService) protected readonly storageService: LocalStorageService,
        @inject(FileSystem) protected readonly fileSystem: FileSystem,
        @inject(WorkspaceService) protected readonly workspaceService: WorkspaceService,
        @inject(GAD) protected readonly gad: GAD,
        @inject(AuthService) protected readonly authService: AuthService,
        @inject(MessageService) protected readonly messageService: MessageService,
        @inject(ProblemManager) protected readonly problemManager: ProblemManager,
    ) { }

    @postConstruct()
    protected async init() {

        this.gitRepositoryTracker.onDidChangeRepository(async repository => {
            if (repository) {
                await this.loadFeature(repository.localUri);
                this.storageService.setData(this.selectedFeatureStorageKey, this._selectedFeature);
                this.fireFeatureChange(this.selectedFeature);
            }
        });

        this.gitRepositoryTracker.onGitEvent(async event => {
            this._selectedFeature = await this.loadFeature(event.source.localUri);

            if (this._selectedFeature) {
                this._selectedFeature.status = event.status;
            }
            this.storageService.setData(this.selectedFeatureStorageKey, this._selectedFeature);
            this.fireFeatureChange(this.selectedFeature);
        }
        );

        this.onFeatureChangeEvent(async feature => {
            document.title = `${feature.name}:${feature.version}`;
        });

        this.workspaceService.onWorkspaceChanged(event => {
            this.refreshFeatures();
        });

        this.refreshFeatures();

        this._selectedFeature = await this.storageService.getData<Feature | undefined>(this.selectedFeatureStorageKey);
        this.fireFeatureChange(this.selectedFeature);

        this.onBuild(async event => {
            if (event.type === 'end') {
                if (event.result === 'ok') {
                    this.messageService.info('Build realizado com sucesso', { timeout: 60000 });
                    this.problemManager.setMarkers(new URI(this.selectedFeature.localUri), 'BUILD', []);
                } else {
                    this.problemManager.setMarkers(new URI(this.selectedFeature.localUri), 'BUILD', [{
                        message: event.details,
                        severity: 1,
                        range: {
                            start: { line: 1, character: 1 },
                            end: { line: 1, character: 1 }
                        },
                    }]);
                }
            }
        });

        this.onDeploy(event => {
            if (event.type === 'end') {
                if (event.result === 'ok') {
                    this.messageService.info(`Deploy realizado com sucesso. ${event.details} `, { timeout: 60000 });
                    this.problemManager.setMarkers(new URI(this.selectedFeature.localUri), 'DEPLOY', []);
                } else {
                    this.problemManager.setMarkers(new URI(this.selectedFeature.localUri), 'DEPLOY', [{
                        message: event.details,
                        severity: 1,
                        range: {
                            start: { line: 1, character: 1 },
                            end: { line: 1, character: 1 }
                        },
                    }]);
                }
            }
        });
    }

    protected async refreshFeatures(): Promise<void> {
        const roots: FileStat[] = [];
        for (const root of await this.workspaceService.roots) {
            if (await this.fileSystem.exists(`${root.uri}/.feature`)) {
                roots.push(root);
            }
        }

        if (roots.length > 0) {
            this._workspaceFeatures = await Promise.all(
                roots.map(r => this.loadFeature(r.uri))
            );
        }
        this.storageService.setData<Feature[]>(this.workspaceFeaturesStorageKey, this._workspaceFeatures);
    }

    protected async loadFeature(uri: string): Promise<Feature> {

        try {
            return this.gad.getFeature(uri);
        } catch (error) {
            console.log(error);
            return new Error(error);
        }
    }

    protected fireFeatureChange(feature: Feature): void {
        this.onFeatureChangeEmitter.fire(feature);
    }

    async lsFeatureVersions(id: string): Promise<FeatureVersion[]> {
        return this.gad.lsFeatureVersions(id);
    }

    async buildFeature(): Promise<void> {

        const feature = this.selectedFeature;
        const user = this.authService.getUserInfo();
        const contextId = feature.context ? feature.context.id : '';

        this.building = true;
        this.onBuildEmitter.fire({
            type: 'start'
        });

        const environments = await this.gad.lsEnvironment(contextId);

        const enviromentDev = environments.find(env => env.type === 'DEVELOPMENT');

        if (enviromentDev) {
            const result = await this.gad.buildFeature(feature, enviromentDev, user, true);

            this.building = false;
            if (result.status === 'ok') {
                this.onBuildEmitter.fire({
                    type: 'end',
                    result: 'ok',
                    details: result.details
                });
            } else {
                this.onBuildEmitter.fire({
                    type: 'end',
                    result: 'nok',
                    details: result.details
                });
            }

        } else {
            this.onBuildEmitter.fire({
                type: 'end',
                result: 'nok',
                details: 'Não encontrado ambiente de desenvolvimento configurado'
            });
        }
    }

    async deployFeature(): Promise<void> {

        const feature = this.selectedFeature;
        const user = this.authService.getUserInfo();
        const contextId = feature.context ? feature.context.id : '';

        this.deploying = true;
        this.onDeployEmitter.fire({
            type: 'start'
        });

        const environments = await this.gad.lsEnvironment(contextId);
        const enviromentDev = environments.find(env => env.type === 'DEVELOPMENT');

        if (enviromentDev) {
            let result = await this.gad.buildFeature(feature, enviromentDev, user);

            if (result.status === 'ok') {
                result = await this.gad.provisionFeature(feature, enviromentDev, user);

                this.deploying = false;
                if (result.status === 'ok') {
                    this.onDeployEmitter.fire({
                        type: 'end',
                        result: 'ok',
                        details: result.details
                    });
                } else {
                    this.onDeployEmitter.fire({
                        type: 'end',
                        result: 'nok',
                        details: result.details
                    });
                }

            } else {
                this.deploying = false;
                this.onDeployEmitter.fire({
                    type: 'end',
                    result: 'nok',
                    details: result.details
                });
            }

        } else {
            this.deploying = false;
            this.onDeployEmitter.fire({
                type: 'end',
                result: 'nok',
                details: 'Não encontrado ambiente de desenvolvimento configurado'
            });
        }
    }

    get onFeatureChangeEvent(): Event<Feature> {
        return this.onFeatureChangeEmitter.event;
    }

    get selectedFeature(): Feature {
        return this._selectedFeature || {};
    }
}
