/********************************************************************************
 * Copyright (C) 2017 TypeFox and others.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied: GNU General Public License, version 2
 * with the GNU Classpath Exception which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 ********************************************************************************/

import { injectable, inject, postConstruct } from 'inversify';
import { GAD, FeatureVersion, Feature, Environment, Result } from '../common';
import { FileSystem } from '@theia/filesystem/lib/common';
import * as jsoncparser from 'jsonc-parser';

const axios = require('axios');

import { User } from '@quarks/auth-extension/lib/common';

@injectable()
export class GADImpl implements GAD {

    protected hosts: any;

    dispose(): void {

    }

    constructor(
        @inject(FileSystem) protected readonly fileSystem: FileSystem,
    ) { }

    @postConstruct()
    protected async init(): Promise<void> {
        this.hosts = (await axios.get(process.env.CONFIG_URI)).data;
    }

    async lsFeatureVersions(id: string): Promise<FeatureVersion[]> {
        const request = {
            featureId: id
        };

        try {
            let result = await axios.post(`${this.hosts.HOST_GAD}/feature/repositoryVersionInfo`, request);

            result = result.data.versions.map((item: any) => (
                {
                    version: item.version,
                    status: item.status.name,
                    description: item.status.description,
                    uri: item.checkoutTo.urlGit,
                    type: item.checkoutTo.type,
                    ref: item.checkoutTo.value,
                    allowNewVersion: result.data.allowNewVersion,
                    allowNewBugfix: result.data.allowNewBugfix
                }
            ));
            return result;
        } catch (error) {
            console.log(error);
            throw new Error(error);
        }

    }

    async getFeature(uri: string): Promise<Feature> {

        try {
            const { content } = await this.fileSystem.resolveContent(`${uri}/.feature`);
            const featureConfig = jsoncparser.parse(content);
            let feature = {
                localUri: uri,
                id: featureConfig.id,
                name: featureConfig.name,
                version: featureConfig.version
            } as Feature;

            const result = await axios.get(`${this.hosts.HOST_GAD_REPOSITORY}/feature/${feature.id}`);

            feature = {
                ...feature,
                context: {
                    id: result.data.context.id
                }
            };

            return feature;
        } catch (error) {
            console.log(error);
            return new Error(error);
        }
    }

    async importFeature(feature: Feature, user: User): Promise<Result> {
        return { status: 'ok' };
    }

    async buildFeature(feature: Feature, environment: Environment, user: User, simulate: boolean = false): Promise<Result> {

        const request = {
            featureId: feature.id,
            featureName: feature.name,
            username: user.username,
            environmentName: environment.name,
            simulate
        };

        try {
            const result = await axios.post(`${this.hosts.HOST_GAD}/builder/build`, request);
            console.log(`buildFeature: ${result}`);

            if (result.data.status === 'error') {
                console.log('buildFeature:');
                console.log(result.data.errorMessage);
                return {
                    status: 'nok',
                    details: result.data.errorMessage
                };
            }

            return {
                status: 'ok',
                details: result.data.message
            };

        } catch (error) {
            console.log('buildFeature:');
            console.log(error);
            return {
                status: 'nok',
                details: error.message
            };
        }
    }

    async provisionFeature(feature: Feature, environment: Environment, user: User): Promise<Result> {
        const request = {
            featureId: feature.id,
            featureName: feature.name,
            username: user.username,
            environmentId: environment.id,
        };

        console.log(request);

        try {
            const result = await axios.post(`${this.hosts.HOST_GAD_FS}/feature/provisionFeature`, request);
            console.log(`provisionFeature: ${result}`);

            if (result.data.status === 'error') {
                console.log('provisionFeature:');
                console.log(result.data.errorMessage);
                return {
                    status: 'nok',
                    details: result.data.errorMessage
                };
            }

            return {
                status: 'ok',
                details: result.data.route
            };

        } catch (error) {
            console.log('provisionFeature catch:');
            console.log(error);
            return {
                status: 'nok',
                details: error.message
            };
        }
    }

    async newFeature(feature: Feature, user: User): Promise<Result> {
        return { status: 'ok' };
    }

    async lsEnvironment(contextId: string): Promise<Environment[]> {

        try {
            const result = await axios.get(`${this.hosts.HOST_GAD_REPOSITORY}/environment/findEnvironmentByContext/${contextId}`);

            const response: Environment[] = result.data.map((element: any) => (
                {
                    id: element.id,
                    name: element.name,
                    type: element.type
                } as Environment
            ));

            return response;

        } catch (error) {
            console.log(error);
            throw new Error(error);
        }
    }
}
